/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "Sensor.h"
#include "Body.h"
#include "Link.h"
#include <cnoid/EigenUtil>

using namespace cnoid;

Sensor::Sensor()
{
    type = COMMON;
}


Sensor* Sensor::create(int type)
{
    Sensor* sensor;

    switch(type){
	
    case FORCE:
        sensor = new ForceSensor();
        break;
		
    case RATE_GYRO:
        sensor = new RateGyroSensor();
        break;
		
    case ACCELERATION:
        sensor = new AccelSensor();
        break;

    case RANGE:
        sensor = new RangeSensor();
        break;

    case PRESSURE:
    case PHOTO_INTERRUPTER:
    case VISION:
    case TORQUE:
        sensor = new Sensor();
        break;
    default:
        sensor = 0;
    }
	
    return sensor;
}


Sensor:: ~Sensor()
{
	
}


void Sensor::operator=(const Sensor& rhs)
{
    name     = rhs.name;
    type     = rhs.type;
    id       = rhs.id;
    localR   = rhs.localR;
    localPos = rhs.localPos;
}
	

void Sensor::destroy(Sensor* sensor)
{
    delete sensor;
}


void Sensor::clear()
{

}


void Sensor::putInformation(std::ostream &os)
{
    os << "name = " << name << ", id = " << id << "\n";
    os << "localAttitude = " << localR << ", localPos = " << localPos << std::endl;
}


ForceSensor::ForceSensor()
{
    type = FORCE;
}


void ForceSensor::clear()
{
    f.setZero();
    tau.setZero();
}


void ForceSensor::putInformation(std::ostream& os)
{
    os << "Force Sensor\n";
    Sensor::putInformation(os);
    os << "f = " << f << "tau = " << tau << std::endl;
}


RateGyroSensor::RateGyroSensor()
{
    type = RATE_GYRO;
}


void RateGyroSensor::clear()
{
    w.setZero();
}


void RateGyroSensor::putInformation(std::ostream& os)
{
    os << "Gyro\n";
    Sensor::putInformation(os);
    os << "omega = " << w << std::endl;
}


AccelSensor::AccelSensor()
{
    type = ACCELERATION;
    clear();
}


void AccelSensor::clear()
{
    dv << 0.0, 0.0, 9.8;
}


void AccelSensor::putInformation(std::ostream& os)
{
    os << "Acceleration Sensor\n";
    Sensor::putInformation(os);
    os << "dv = " << dv << std::endl;
}

RangeSensor::RangeSensor()
{
    type = RANGE;
    scanAngle = PI;
    scanStep = 0.1;
    scanRate = 10;
    maxDistance = 10;
    nextUpdateTime = 0;
}


namespace cnoid {

    class SensorCalculatorImpl
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SensorCalculator::ForceFunction forceFunction;
        
        // preview control gain matrices for acceleration sensors
        Matrix2 A;
        Vector2 B;

        Vector3 g;

        bool isInitialAccelSensorCalculation;

        void initialize(double timeStep, const Vector3& gravityAcceleration) {

            isInitialAccelSensorCalculation = true;
            g = gravityAcceleration;

            // Kalman filter design
            static const double n_input = 100.0;  // [N]
            static const double n_output = 0.001; // [m/s]

            // Analytical solution of Kalman filter (continuous domain)
            // s.kajita  2003 Jan.22

            Matrix2 Ac;
            Ac <<
                -sqrt(2.0 * n_input / n_output), 1.0,
                -n_input / n_output            , 0.0;

            Vector2 Bc(sqrt(2.0 * n_input / n_output), n_input / n_output);

            A.setIdentity();
            Matrix2 An = Matrix2::Identity();
            Matrix2 An2;
            B = timeStep * Bc;
            Vector2 Bn = B;
            Vector2 Bn2;

            double factorial[14];
            double r = 1.0;
            factorial[1] = r;
            for(int i=2; i <= 13; ++i){
                r += 1.0;
                factorial[i] = factorial[i-1] * r;
            }

            for(int i=1; i <= 12; i++){
                An2.noalias() = Ac * An;
                An = timeStep * An2;
                A += (1.0 / factorial[i]) * An;
                
                Bn2.noalias() = Ac * Bn;
                Bn = timeStep * Bn2;
                B += (1.0 / factorial[i+1]) * Bn;
            }
        }

        void updateAccelSensor(AccelSensor* sensor) {

            Link* link = sensor->link;
            Vector2* x = sensor->x;

            Vector3 o_Vgsens = link->R * (link->R.transpose() * link->w).cross(sensor->localPos) + link->v;

            if(isInitialAccelSensorCalculation){
                for(int i=0; i < 3; ++i){
                    x[i](0) = o_Vgsens(i);
                    x[i](1) = 0.0;
                }
            } else {
                // kalman filtering
                for(int i=0; i < 3; ++i){
                    x[i] = A * x[i] + o_Vgsens(i) * B;
                }
            }

            Vector3 o_Agsens(x[0](1), x[1](1), x[2](1));
            o_Agsens -= g;
            
            sensor->dv.noalias() = link->R.transpose() * o_Agsens;
        }
        
    };
}


SensorCalculator::SensorCalculator()
{
    impl = new SensorCalculatorImpl();
}


SensorCalculator::~SensorCalculator()
{
    delete impl;
}

    
void SensorCalculator::setForceFunction(ForceFunction func)
{
    impl->forceFunction = func;
}


void SensorCalculator::initialize(double timeStep, const Vector3& gravityAcceleration)
{
    impl->initialize(timeStep, gravityAcceleration);
}


void SensorCalculator::calcForceSensorValues(Body* body, Referenced* userData)
{
    int n = body->numSensors(Sensor::FORCE);

    for(int i=0; i < n; ++i){

        ForceSensor* sensor = body->sensor<ForceSensor>(i);
        Link* link = sensor->link;

        Vector6 wrench;
        impl->forceFunction(sensor, wrench, userData);
        
        const Matrix3 sensorR = link->R * sensor->localR;
        const Vector3 sensorPos = link->p + link->R * sensor->localPos;
        
        sensor->f   = sensorR.transpose() * wrench.head<3>();
        sensor->tau = sensorR.transpose() * (wrench.tail<3>() - sensorPos.cross(wrench.head<3>()));
    }
}


void SensorCalculator::calcGyroAndAccelSensorValues(Body* body)
{
    int n = body->numSensors(Sensor::RATE_GYRO);
    for(int i=0; i < n; ++i){
        RateGyroSensor* sensor = body->sensor<RateGyroSensor>(i);
        Link* link = sensor->link;
        sensor->w = sensor->localR.transpose() * link->R.transpose() * link->w;
    }

    n = body->numSensors(Sensor::ACCELERATION);
    for(int i=0; i < n; ++i){
        impl->updateAccelSensor(body->sensor<AccelSensor>(i));
    }

    impl->isInitialAccelSensorCalculation = false;
}
