/**
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_UTIL_YAML_NODES_H_INCLUDED
#define CNOID_UTIL_YAML_NODES_H_INCLUDED

#include "UTF8.h"
#include <map>
#include <vector>
#include <cstring>
#include <iosfwd>
#include <boost/intrusive_ptr.hpp>
#include <boost/function.hpp>
#include "exportdecl.h"

namespace cnoid {

    class YAMLReaderImpl;
    class YAMLWriter;

    class ValueNode;

    void intrusive_ptr_add_ref(cnoid::ValueNode* obj);
    void intrusive_ptr_release(cnoid::ValueNode* obj);

    class ScalarNode;
    class Mapping;
    class Sequence;

#ifndef CNOID_BACKWARD_COMPATIBILITY
    enum StringStyle { PLAIN_STRING, SINGLE_QUOTED, DOUBLE_QUOTED, LITERAL_STRING, FOLDED_STRING };
#else
    enum StringStyle { PLAIN_STRING, YAML_PLAIN_STRING = PLAIN_STRING,
                       SINGLE_QUOTED, YAML_SINGLE_QUOTED = SINGLE_QUOTED,
                       DOUBLE_QUOTED, YAML_DOUBLE_QUOTED = DOUBLE_QUOTED,
                       LITERAL_STRING, YAML_LITERAL = LITERAL_STRING,
                       FOLDED_STRING, YAML_FOLDED = FOLDED_STRING
    };
#endif

    class CNOID_EXPORT ValueNode
    {
        struct Initializer {
            Initializer();
        };
        static Initializer initializer;
        
      public:

        enum Type { INVALID_NODE = 0, MAPPING, SEQUENCE, SCALAR, LF_NODE };

        inline bool isValid() const { return type_ != INVALID_NODE; }

        inline Type type() const { return type_; }

        int toInt() const;
        double toDouble() const;
        bool toBool() const;

        inline bool isString() const { return type_ == SCALAR; }

#ifdef _WIN32
        const std::string toString() const;
        const std::string toUTF8String() const;

        inline operator std::string () const {
            return toString();
        }
#else
        const std::string& toString() const;
        const std::string& toUTF8String() const;

        inline operator const std::string& () const {
            return toString();
        }
#endif

        inline bool isMapping() const { return type_ == MAPPING; }
        const Mapping* toMapping() const;
        Mapping* toMapping();

        inline bool isSequence() const { return type_ == SEQUENCE; }
        const Sequence* toSequence() const;
        Sequence* toSequence();

        //bool read(std::string &out_value) const;
        //bool read(bool &out_value) const;
        bool read(int &out_value) const;
        //bool read(double &out_value) const;

        inline bool hasLineInfo() const { return (line_ >= 0); }
        inline int line() const { return line_ + 1; }
        inline int column() const { return column_ + 1; }

        void throwException(const std::string& message) const;

        class CNOID_EXPORT Exception {
        public:
            virtual ~Exception();
            int line() const { return line_; }
            int column() const { return column_; }
            const std::string& message() const { return message_; }
            void setPosition(int line, int column) {
                line_ = line;
                column_ = column;
            }
            void setMessage(const std::string& m){
                message_ = m;
            }
          private:
            int line_;
            int column_;
            std::string message_;
        };
        
        class KeyNotFoundException : public Exception {
        public:
            const std::string& key() { return key_; }
            void setKey(const std::string& key) { key_ = key; }
        private:
            std::string key_;
        };
        
        class NotScalarException : public Exception {
        };
        
        class ScalarTypeMismatchException : public Exception {
        };

        class NotMappingException : public Exception {
        };

        class NotSequenceException : public Exception {
        };

        class SyntaxException : public Exception {
        };

        class DocumentNotFoundException : public Exception {
        };

      private:
        
        int refCounter;
        
      protected:

        ValueNode() : refCounter(0) { }
        ValueNode(Type type) : refCounter(0), type_(type), line_(-1), column_(-1) { }

        virtual ~ValueNode() { }

        void throwNotScalrException() const;
        void throwNotMappingException() const;
        void throwNotSequenceException() const;

        Type type_;

      private:

        // disabled copy operations
        ValueNode(const ValueNode&);
        ValueNode& operator=(const ValueNode&);

        int line_;
        int column_;
        int indexInMapping; // used for YAMLWriter

        friend class YAMLReaderImpl;
        friend class YAMLWriter;
        friend class ScalarNode;
        friend class Mapping;
        friend class Sequence;

        friend void cnoid::intrusive_ptr_add_ref(ValueNode* obj);
        friend void cnoid::intrusive_ptr_release(ValueNode* obj);
    };

    typedef boost::intrusive_ptr<ValueNode> ValueNodePtr;

    
    class CNOID_EXPORT ScalarNode : public ValueNode
    {
      private:
        ScalarNode(const char* text, size_t length);
        ScalarNode(const char* text, size_t length, StringStyle stringStyle);
        ScalarNode(const std::string& value, StringStyle stringStyle);

        std::string stringValue;
        StringStyle stringStyle;

        friend class YAMLReaderImpl;
        friend class YAMLWriter;
        friend class ValueNode;
        friend class Mapping;
        friend class Sequence;
    };


    class CNOID_EXPORT Collection : public ValueNode
    {
      public:
        virtual ~Collection();

        inline void setFlowStyle(bool isFlowStyle = true) { isFlowStyle_ = isFlowStyle; }
        inline bool isFlowStyle() const { return isFlowStyle_; }

        void setDoubleFormat(const char* format);
        inline const char* doubleFormat() { return doubleFormat_; }

      protected:
        Collection();
        const char* doubleFormat_;

      private:
        Collection(const Collection&);
        Collection& operator=(const Collection&);

        bool isFlowStyle_;
    };


    class CNOID_EXPORT Mapping : public Collection
    {
        typedef std::map<std::string, ValueNodePtr> Container;
        
      public:

        typedef Container::iterator iterator;
        typedef Container::const_iterator const_iterator;

        Mapping();
        Mapping(int line, int column);
        virtual ~Mapping();

        inline bool empty() const { return values.empty(); }
        inline size_t size() const { return values.size(); }
        void clear();

        void setKeyQuoteStyle(StringStyle style);

        ValueNode* find(const std::string& key) const;
        Mapping* findMapping(const std::string& key) const;
        Sequence* findSequence(const std::string& key) const;

        ValueNode& get(const std::string& key) const;

        inline ValueNode& operator[](const std::string& key) const {
            return get(key);
        }

        void insert(const std::string& key, ValueNodePtr node);

        inline Mapping* openMapping(const std::string& key) {
            return openMapping(key, false);
        }
        
        inline Mapping* openFlowStyleMapping(const std::string& key) {
            return openFlowStyleMapping(key, false);
        }

        inline Mapping* createMapping(const std::string& key) {
            return openMapping(key, true);
        }
        
        inline Mapping* createFlowStyleMapping(const std::string& key) {
            return openFlowStyleMapping(key, true);
        }

        inline Sequence* openSequence(const std::string& key) {
            return openSequence(key, false);
        }
        
        inline Sequence* openFlowStyleSequence(const std::string& key){
            return openFlowStyleSequence(key, false);
        }

        inline Sequence* createSequence(const std::string& key){
            return openSequence(key, true);
        }
        
        inline Sequence* createFlowStyleSequence(const std::string& key){
            return openFlowStyleSequence(key, true);
        }
        
        bool read(const std::string &key, std::string &out_value) const;
        bool readUTF8(const std::string &key, std::string &out_value) const;
        bool read(const std::string &key, bool &out_value) const;
        bool read(const std::string &key, int &out_value) const;
        bool read(const std::string &key, double &out_value) const;

        bool read(const std::string &key, boost::function<void(double)> setterFunc) const;

        
        template <class T>
        inline T read(const std::string& key) const {
            T value;
            if(read(key, value)){
                return value;
            } else {
                throwKeyNotFoundException(key);
            }
        }

        template <class T>
        inline T get(const std::string& key, const T& defaultValue) const {
            T value;
            if(read(key, value)){
                return value;
            } else {
                return defaultValue;
            }
        }

        inline std::string get(const std::string& key, const char* defaultValue) const {
            std::string value;
            if(read(key, value)){
                return value;
            } else {
                return defaultValue;
            }
        }

        void writeUTF8(const std::string &key, const std::string& value, StringStyle stringStyle = PLAIN_STRING);

        void write(const std::string &key, const std::string& value, StringStyle stringStyle = PLAIN_STRING) {
            writeUTF8(key, toUTF8(value), stringStyle);
        }

        void writeUTF8(const std::string &key, const char* value, StringStyle stringStyle = PLAIN_STRING){
            writeUTF8(key, std::string(value), stringStyle);
        }
        
        void write(const std::string &key, const char* value, StringStyle stringStyle = PLAIN_STRING){
            write(key, std::string(value), stringStyle);
        }

        void write(const std::string &key, bool value);
        void write(const std::string &key, int value);
        void write(const std::string &key, double value);
        void writePath(const std::string &key, const std::string& value);

        typedef enum { READ_MODE, WRITE_MODE } AssignMode;

        inline void setAssignMode(AssignMode mode) { this->mode = mode; }

        template <class T>
        void assign(const std::string& key, T& io_value, const T& defaultValue){
            switch(mode){
            case READ_MODE:
                if(!read(key, io_value)){
                    io_value = defaultValue;
                }
                break;
            case WRITE_MODE:
                write(key, io_value);
                break;
            }
        }
        
        inline iterator begin() { return values.begin(); }
        inline iterator end() { return values.end(); }
        inline const_iterator begin() const { return values.begin(); }
        inline const_iterator end() const { return values.end(); }

        void throwKeyNotFoundException(const std::string& key) const;

    private:

        Mapping(const Mapping&);
        Mapping& operator=(const Mapping&);

        Mapping* openMapping(const std::string& key, bool doOverwrite);
        Mapping* openFlowStyleMapping(const std::string& key, bool doOverwrite);
        Sequence* openSequence(const std::string& key, bool doOverwrite);
        Sequence* openFlowStyleSequence(const std::string& key, bool doOverwrite);

        inline void insertSub(const std::string& key, ValueNode* node);

        void writeSub(const std::string &key, const char* text, size_t length, StringStyle stringStyle);

        static bool compareIters(const Mapping::const_iterator& it1, const Mapping::const_iterator& it2);

        Container values;
        AssignMode mode;
        int indexCounter;
        StringStyle keyQuoteStyle;

        friend class Sequence;
        friend class YAMLReaderImpl;
        friend class YAMLWriter;
    };

    typedef boost::intrusive_ptr<Mapping> MappingPtr;


    /**
       @todo add 'openMapping' and 'openSequence' methods
    */
    class CNOID_EXPORT Sequence : public Collection
    {
        typedef std::vector<ValueNodePtr> Container;

    public:

        Sequence();
        Sequence(int size);
        ~Sequence();
        
        typedef Container::iterator iterator;
        typedef Container::const_iterator const_iterator;

        inline bool empty() const { return values.empty(); }
        inline int size() const { return values.size(); }
        void clear();
        void reserve(int size);

        inline ValueNode& front() const {
            return *values.front();
        }

        inline ValueNode& back() const {
            return *values.back();
        }

        ValueNode* at(int i) const;

        inline ValueNode& get(int i) const {
            return *values[i];
        }

        void write(int i, int value);
        void write(int i, const std::string& value, StringStyle stringStyle = PLAIN_STRING);

        bool read(int i, bool &out_value) const;
        bool read(int i, int &out_value) const;
        bool read(int i, double &out_value) const;

        inline ValueNode& operator[](int i) const {
            return *values[i];
        }

        /// \todo implemente the following funcion (ticket #35)
        //MappingPtr extractMapping(const std::string& key) const;

        Mapping* newMapping();

	inline void append(ValueNodePtr node) {
            values.push_back(node);
        }
        
        void append(int value);

        /**
            @param maxColumns LF is automatically inserted when the column pos is over maxColumsn
            @param numValues If numValues is not greater than maxColumns, the initial LF is skipped.
                   This feature is disabled if numValues = 0.
         */
        inline void append(int value, int maxColumns, int numValues = 0) {
            insertLF(maxColumns, numValues);
            append(value);
        }

        void append(size_t value);

        /**
           @param maxColumns LF is automatically inserted when the column pos is over maxColumsn
           @param numValues If numValues is not greater than maxColumns, the initial LF is skipped.
                  This feature is disabled if numValues = 0.
        */
        /*
        inline void append(size_t value, int maxColumns, int numValues = 0){
            insertLF(maxColumns, numValues);
            append(value);
        }
        */
        
        void append(double value);

        /**
           @param maxColumns LF is automatically inserted when the column pos is over maxColumsn
           @param numValues If numValues is not greater than maxColumns, the initial LF is skipped.
                  This feature is disabled if numValues = 0.
        */
        inline void append(double value, int maxColumns, int numValues = 0) {
            insertLF(maxColumns, numValues);
            append(value);
        }

        void append(const std::string& value, StringStyle stringStyle = PLAIN_STRING);

        /**
           @param maxColumns LF is automatically inserted when the column pos is over maxColumsn
           @param numValues If numValues is not greater than maxColumns, the initial LF is skipped.
                  This feature is disabled if numValues = 0.
        */
        inline void append(const std::string& value, int maxColumns, int numValues = 0, StringStyle stringStyle = PLAIN_STRING){
            insertLF(maxColumns, numValues);
            append(value, stringStyle);
        }

        void appendLF();
        
        inline iterator begin() { return values.begin(); }
        inline iterator end() { return values.end(); }
        inline const_iterator begin() const { return values.begin(); }
        inline const_iterator end() const { return values.end(); };

    private:

        Sequence(int line, int column);
        Sequence(int line, int column, int reservedSize);
        
        Sequence(const Sequence&);
        Sequence& operator=(const Sequence&);

        void insertLF(int maxColumns, int numValues);
        
        Container values;

        friend class Mapping;
        friend class YAMLReaderImpl;
        friend class YAMLWriter;
    };

    typedef boost::intrusive_ptr<Sequence> SequencePtr;

    inline void intrusive_ptr_add_ref(cnoid::ValueNode* obj){
        obj->refCounter++;
    }

    inline void intrusive_ptr_release(cnoid::ValueNode* obj){
        obj->refCounter--;
        if(obj->refCounter == 0){
            delete obj;
        }
    }

#ifdef CNOID_BACKWARD_COMPATIBILITY
    typedef ValueNode YamlNode;
    typedef ValueNodePtr YamlNodePtr;
    typedef ScalarNode YamlScalar;
    typedef Collection YamlCollection;
    typedef Mapping YamlMapping;
    typedef MappingPtr YamlMappingPtr;
    typedef Sequence YamlSequence;
    typedef SequencePtr YamlSequencePtr;
#endif
}

#endif
